# w.c.s. - web application for online forms
# Copyright (C) 2005-2022  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from quixote import get_publisher

from wcs import sql
from wcs.fields import Field

from .qommon import _, misc
from .qommon.storage import Equal


class TestError(Exception):
    pass


class TestDef(sql.TestDef):
    _names = 'testdef'

    name = ''
    slug = None
    object_type = None  # (formdef, carddef, etc.)
    object_id = None

    data = None  # (json export of formdata, carddata, etc.)

    def __str__(self):
        return self.name

    def store(self, *args, comment=None, snapshot_store_user=True, **kwargs):
        if not self.slug:
            existing_slugs = {
                x.slug
                for x in self.select(
                    [Equal('object_type', self.object_type), Equal('object_id', self.object_id)]
                )
            }
            base_slug = misc.simplify(self.name)
            self.slug = base_slug
            i = 2
            while self.slug in existing_slugs:
                self.slug = '%s-%s' % (base_slug, i)
                i += 1
        super().store(*args, **kwargs)

    @classmethod
    def create_from_formdata(cls, formdef, formdata):
        testdef = cls()
        testdef.object_type = formdef.get_table_name()
        testdef.object_id = formdef.id
        testdef.data = {
            'fields': formdata.data,
            'user': formdata.user.get_json_export_dict() if formdata.user else None,
        }
        return testdef

    def run(self, objectdef):
        formdata = objectdef.data_class()()
        if self.data['user']:
            formdata.set_user_from_json(self.data['user'])

        get_publisher().substitutions.reset()
        get_publisher().substitutions.feed(get_publisher())
        get_publisher().substitutions.feed(objectdef)
        get_publisher().substitutions.feed(formdata)

        pages = [field for field in objectdef.fields if field.type == 'page']
        self.page_indexes = {page: i for i, page in enumerate(pages, 1)}

        current_page = None
        for field in objectdef.fields:
            if field.type == 'page':
                if current_page:  # evaluate previous page post conditions
                    self.evaluate_page_conditions(current_page, formdata, objectdef)
                current_page = field
                continue

            if field.type in ('subtitle', 'title', 'comment', 'computed'):
                continue

            if not field.is_visible(formdata.data, objectdef):
                if field.id in self.data['fields']:
                    raise TestError(
                        _('Tried to fill field "%(label)s" on page %(no)d but it is hidden.')
                        % {'label': field.label, 'no': self.page_indexes.get(current_page, 1)}
                    )
                continue

            formdata.data[field.id] = self.data['fields'].get(field.id)
            get_publisher().substitutions.invalidate_cache()

        if current_page:  # evaluate last page post conditions
            self.evaluate_page_conditions(current_page, formdata, objectdef)

    def evaluate_page_conditions(self, page, formdata, objectdef):
        for post_condition in page.post_conditions or []:
            condition = post_condition.get('condition', {})
            try:
                if not Field.evaluate_condition(formdata.data, objectdef, condition):
                    raise TestError(
                        _('Page %(no)d post condition was not met (%(condition)s).')
                        % {'no': self.page_indexes[page], 'condition': condition.get('value')}
                    )
            except RuntimeError:
                raise TestError(_('Failed to evaluate page %d post condition.') % self.page_indexes[page])

    def export_to_json(self):
        return {
            'name': self.name,
            'slug': self.slug,
            'object_type': self.object_type,
            'object_id': self.object_id,
            'data': self.data,
        }

    @classmethod
    def import_from_json(cls, data):
        testdefs = TestDef.select(
            [
                Equal('object_type', data['object_type']),
                Equal('object_id', data['object_id']),
                Equal('slug', data['slug']),
            ]
        )

        testdef = testdefs[0] if testdefs else TestDef()

        for k, v in data.items():
            setattr(testdef, k, v)

        testdef.store()
        return testdef
